from typing import Tuple, List, Set, Union, cast, Dict, Optional, Any

import torch
from torch import Tensor, nn

from deepxube.utils import misc_utils
from deepxube.nnet.pytorch_models import FullyConnectedModel, ResnetModel, Conv2dModel

import numpy as np
from numpy.typing import NDArray
import matplotlib.pyplot as plt
from deepxube.environments.environment_abstract import EnvGrndAtoms, State, Action, HeurFnNNet, Goal
from deepxube.logic.logic_objects import Atom, Model

import pickle

import re

import pathlib
import tarfile
import os
import wget  # type: ignore
from filelock import FileLock


class NNet(HeurFnNNet):
    def __init__(self, h1_dim: int, resnet_dim: int, num_resnet_blocks: int, out_dim: int, batch_norm: bool,
                 nnet_type: str):
        super().__init__(nnet_type)
        # self.img_nnet = Conv2dModel(3, [16, 16], [2, 2], [0, 0], [True, True], ["RELU", "RELU"], strides=[2, 2])

        self.conv_to_flat = nn.Sequential(
            Conv2dModel(6, [16, 16], [2, 2], [1, 0], [True, True], ["RELU", "RELU"], strides=[1, 1]),
            nn.Flatten(),
        )

        self.first_fc = FullyConnectedModel(1600, [h1_dim, resnet_dim], [batch_norm] * 2, ["RELU"] * 2)
        self.resnet = ResnetModel(resnet_dim, num_resnet_blocks, out_dim, batch_norm, layer_act="RELU")

    def forward(self, states_goals_l: List[Tensor]):
        # img_nnet_out: Tensor = self.img_nnet(states_l[0])
        x = self.conv_to_flat(torch.cat((states_goals_l[0], states_goals_l[1]), dim=1).float())
        x = self.first_fc(x)
        x = self.resnet(x)

        return x


class SokobanState(State):
    __slots__ = ['agent', 'walls', 'boxes', 'hash']

    def __init__(self, agent: NDArray[np.int_], boxes: NDArray[np.uint8], walls: NDArray[np.uint8]):
        self.agent: NDArray[np.int_] = agent
        self.boxes: NDArray[np.uint8] = boxes
        self.walls: NDArray[np.uint8] = walls

        self.hash: Optional[int] = None

    def __hash__(self):
        if self.hash is None:
            boxes_flat = self.boxes.flatten()
            walls_flat = self.walls.flatten()
            state: NDArray[np.int_] = np.concatenate((self.agent, boxes_flat.astype(int),
                                                      walls_flat.astype(int)), axis=0)

            self.hash = hash(state.tobytes())

        return self.hash

    def __eq__(self, other: object):
        if isinstance(other, SokobanState):
            agents_eq: bool = np.array_equal(self.agent, other.agent)
            boxes_eq: bool = np.array_equal(self.boxes, other.boxes)
            walls_eq: bool = np.array_equal(self.walls, other.walls)

            return agents_eq and boxes_eq and walls_eq
        return NotImplemented


class SokobanGoal(Goal):
    __slots__ = ['agent', 'walls', 'boxes']

    def __init__(self, agent: NDArray[np.int_], boxes: NDArray[np.uint8], walls: NDArray[np.uint8]):
        self.agent: NDArray[np.int_] = agent
        self.boxes: NDArray[np.uint8] = boxes
        self.walls: NDArray[np.uint8] = walls


class SkAction(Action):
    def __init__(self, action: int):
        self.action = action


def load_states(data_dir: str) -> List[SokobanState]:
    states_np = pickle.load(open(f"{data_dir}/sokoban/train.pkl", "rb"))
    # t_file = tarfile.open(file_name, "r:gz")
    # states_np = pickle.load(t_file.extractfile(t_file.getmembers()[1]))

    states: List[SokobanState] = []

    agent_idxs = np.where(states_np == 1)
    box_masks = states_np == 2
    wall_masks = states_np == 4

    for idx in range(states_np.shape[0]):
        agent_idx = np.array([agent_idxs[1][idx], agent_idxs[2][idx]], dtype=int)

        states.append(SokobanState(agent_idx, box_masks[idx], wall_masks[idx]))

    return states


def _get_surfaces() -> List[Any]:
    import imageio.v2 as imageio
    data_dir = get_data_dir()
    img_dir = f"{data_dir}/sokoban/"

    lock = FileLock(f"{data_dir}/file.lock")
    with lock:
        # Load images, representing the corresponding situation
        box = imageio.imread(f"{img_dir}/surface/box.png")
        # box_on_target = imageio.imread(f"{img_dir}/surface/box_on_target.png")
        # box_target = imageio.imread(f"{img_dir}/surface/box_target.png")
        floor = imageio.imread(f"{img_dir}/surface/floor.png")
        player = imageio.imread(f"{img_dir}/surface/player.png")
        # player_on_target = imageio.imread(f"{img_dir}/surface/player_on_target.png")
        wall = imageio.imread(f"{img_dir}/surface/wall.png")

    # surfaces = [wall, floor, box_target, player, box, player_on_target, box_on_target]
    surfaces = [wall, floor, player, box]

    return surfaces


def _get_train_states() -> List[SokobanState]:
    data_dir = get_data_dir()
    lock = FileLock(f"{data_dir}/file.lock")

    with lock:
        states_train: List[SokobanState] = load_states(data_dir)

    return states_train


def _np_to_model(agent: NDArray[np.int_], boxes: NDArray[np.uint8], walls: NDArray[np.uint8]) -> Model:
    grnd_atoms: List[Atom] = []
    if agent.shape[0] > 0:
        grnd_atoms += [('agent', str(int(agent[0])), str(int(agent[1])))]
    grnd_atoms += [('box', str(int(x)), str(int(y))) for x, y in zip(*np.where(boxes))]
    grnd_atoms += [('wall', str(int(x)), str(int(y))) for x, y in zip(*np.where(walls))]

    return frozenset(grnd_atoms)


def get_data_dir() -> str:
    parent_dir: str = str(pathlib.Path(__file__).parent.resolve())
    data_dir: str = f"{parent_dir}/data/sokoban/"

    return data_dir


class Sokoban(EnvGrndAtoms[SokobanState, SkAction, SokobanGoal]):

    def __init__(self, env_name: str):
        super().__init__(env_name)

        self.dim: int = 10
        self.num_boxes: int = 4

        self.num_actions: int = 4

        self.img_dim: int = 160

        self.states_train: Optional[List[SokobanState]] = None
        self._surfaces: Optional[List[Any]] = None

        # check if data needs to be downloaded
        data_dir = get_data_dir()
        data_download_link: str = "https://github.com/forestagostinelli/DeepXubeData/raw/main/sokoban.tar.gz"
        if not os.path.exists(f"{data_dir}/sokoban/"):
            valid_user_in: bool = False
            while not valid_user_in:
                user_in: str = input(f"Sokoban data needs to be downloaded from {data_download_link}. "
                                     f"Download data (about 16MB)? (y/n):")
                if user_in.upper() == "Y":
                    valid_user_in = True
                    print("Downloading compressed data")
                    if not os.path.exists(data_dir):
                        os.makedirs(data_dir)

                    tar_gz_file_name = f"{data_dir}/sokoban.tar.gz"
                    wget.download(data_download_link, tar_gz_file_name, bar=None)
                    tar_gz_file = tarfile.open(tar_gz_file_name)
                    print("Uncompressing data")
                    tar_gz_file.extractall(data_dir)
                    print("Deleting compressed data")
                    os.remove(tar_gz_file_name)
                elif user_in.upper() == "N":
                    valid_user_in = True

    def next_state(self, states: List[SokobanState], actions: List[SkAction]) -> Tuple[List[SokobanState], List[float]]:
        agent = np.stack([state.agent for state in states], axis=0)
        boxes = np.stack([state.boxes for state in states], axis=0)
        walls_next = np.stack([state.walls for state in states], axis=0)

        idxs_arange = np.arange(0, len(states))
        agent_next_tmp = self._get_next_idx(agent, actions)
        agent_next = np.zeros(agent_next_tmp.shape, dtype=int)

        boxes_next = boxes.copy()

        # agent -> wall
        agent_wall = walls_next[idxs_arange, agent_next_tmp[:, 0], agent_next_tmp[:, 1]]
        agent_next[agent_wall] = agent[agent_wall]

        # agent -> box
        agent_box = boxes[idxs_arange, agent_next_tmp[:, 0], agent_next_tmp[:, 1]]
        boxes_next_tmp = self._get_next_idx(agent_next_tmp, actions)

        box_wall = walls_next[idxs_arange, boxes_next_tmp[:, 0], boxes_next_tmp[:, 1]]
        box_box = boxes[idxs_arange, boxes_next_tmp[:, 0], boxes_next_tmp[:, 1]]

        # agent -> box -> obstacle
        agent_box_obstacle = agent_box & (box_wall | box_box)
        agent_next[agent_box_obstacle] = agent[agent_box_obstacle]

        # agent -> box -> empty
        agent_box_empty = agent_box & np.logical_not(box_wall | box_box)
        agent_next[agent_box_empty] = agent_next_tmp[agent_box_empty]
        abe_idxs = np.where(agent_box_empty)[0]

        agent_next_idxs_abe = agent_next[agent_box_empty]
        boxes_next_idxs_abe = boxes_next_tmp[agent_box_empty]

        boxes_next[abe_idxs, agent_next_idxs_abe[:, 0], agent_next_idxs_abe[:, 1]] = False
        boxes_next[abe_idxs, boxes_next_idxs_abe[:, 0], boxes_next_idxs_abe[:, 1]] = True

        # agent -> empty
        agent_empty = np.logical_not(agent_wall | agent_box)
        agent_next[agent_empty] = agent_next_tmp[agent_empty]
        boxes_next[agent_empty] = boxes[agent_empty]

        states_next: List[SokobanState] = []
        for idx in range(len(states)):
            state_next: SokobanState = SokobanState(agent_next[idx], boxes_next[idx], walls_next[idx])
            states_next.append(state_next)

        transition_costs: List[float] = [1.0 for _ in range(len(states))]

        return states_next, transition_costs

    def get_state_actions(self, states: List[SokobanState]) -> List[List[SkAction]]:
        return [[SkAction(x) for x in range(self.num_actions)] for _ in range(len(states))]

    def is_solved(self, states: List[SokobanState], goals: List[SokobanGoal]) -> List[bool]:
        nnet_rep: List[NDArray[np.uint8]] = self.states_goals_to_nnet_input(states, goals)
        states_np: NDArray[np.uint8] = nnet_rep[0].reshape((len(states), -1))
        goals_np: NDArray[np.uint8] = nnet_rep[1].reshape((len(goals), -1))
        is_solved_np = np.all(np.logical_or(states_np == goals_np, goals_np == 0), axis=1)
        return list(is_solved_np)

    def get_start_states(self, num_states: int) -> List[SokobanState]:
        if self.states_train is None:
            self.states_train = _get_train_states()
        state_idxs = np.random.randint(0, len(self.states_train), size=num_states)
        states: List[SokobanState] = [self.states_train[idx] for idx in state_idxs]

        step_range: Tuple[int, int] = (0, 100)

        # Initialize
        scrambs: List[int] = list(range(step_range[0], step_range[1] + 1))

        # Scrambles
        step_nums: NDArray[np.int_] = np.random.choice(scrambs, num_states)
        step_nums_curr: NDArray[np.int_] = np.zeros(num_states, dtype=int)

        # Go backward from goal state
        steps_lt: NDArray[np.bool_] = step_nums_curr < step_nums
        while np.any(steps_lt):
            idxs: NDArray[np.int_] = np.where(steps_lt)[0]

            states_to_move: List[SokobanState] = [states[idx] for idx in idxs]
            actions: List[SkAction] = [SkAction(act) for act in
                                       np.random.randint(0, self.num_actions, size=len(states_to_move))]

            states_moved, _ = self.next_state(states_to_move, actions)

            for idx_moved, idx in enumerate(idxs):
                states[idx] = states_moved[idx_moved]

            step_nums_curr[idxs] = step_nums_curr[idxs] + 1
            steps_lt[idxs] = step_nums_curr[idxs] < step_nums[idxs]

        return states

    def states_goals_to_nnet_input(self, states: List[SokobanState],
                                   goals: List[SokobanGoal]) -> List[NDArray[np.uint8]]:
        """
        states_real: np.ndarray = np.zeros((len(states), self.img_dim, self.img_dim, 3))
        for state_idx, state in enumerate(states):
            states_real[state_idx] = self.state_to_rgb(state)

        states_rep = [states_real.transpose([0, 3, 1, 2])]
        """

        nnet_rep: List[NDArray[np.uint8]] = [self._states_to_np(states), self._goals_to_np(goals)]

        return nnet_rep

    def state_to_model(self, states: List[SokobanState]) -> List[Model]:
        agents_np = np.stack([state.agent for state in states], axis=0)
        boxes_np = np.stack([state.boxes for state in states], axis=0)
        walls_np = np.stack([state.walls for state in states], axis=0)

        models_s: List[Model] = []
        for agent_i, boxes_i, walls_i in zip(agents_np, boxes_np, walls_np):
            models_s.append(_np_to_model(agent_i, boxes_i, walls_i))

        return models_s

    def model_to_state(self, models: List[Model]) -> List[SokobanState]:
        raise NotImplementedError

    def goal_to_model(self, goals: List[SokobanGoal]) -> List[Model]:
        agent = np.stack([goal.agent for goal in goals], axis=0)
        boxes = np.stack([goal.boxes for goal in goals], axis=0)
        walls = np.stack([goal.walls for goal in goals], axis=0)

        models_s: List[Model] = []
        for agent_i, boxes_i, walls_i in zip(agent, boxes, walls):
            models_s.append(_np_to_model(agent_i, boxes_i, walls_i))

        return models_s

    def model_to_goal(self, models: List[Model]) -> List[SokobanGoal]:
        models_np: NDArray[np.uint8] = self._models_to_np(models)
        goals: List[SokobanGoal] = []
        for i in range(len(models)):
            agent_idxs = np.concatenate(np.where(models_np[i, 0]))
            goals.append(SokobanGoal(agent_idxs, models_np[i, 1], models_np[i, 2]))

        return goals

    def get_pddl_domain(self) -> List[str]:
        pddl_str: str = """
        (define (domain sokoban)
  (:requirements :typing )
  (:types thing location direction)
  (:predicates (move-dir ?v0 - location ?v1 - location ?v2 - direction)
    (is-nongoal ?v0 - location)
    (clear ?v0 - location)
    (is-stone ?v0 - thing)
    (at ?v0 - thing ?v1 - location)
    (is-player ?v0 - thing)
    (at-goal ?v0 - thing)
    (move ?v0 - direction)
    (is-goal ?v0 - location)
  )

    (:action move
    :parameters (?p - thing ?from - location ?to - location ?dir - direction)
    :precondition (and (move ?dir)
        (is-player ?p)
        (at ?p ?from)
        (clear ?to)
        (move-dir ?from ?to ?dir))
    :effect (and
        (not (at ?p ?from))
        (not (clear ?to))
        (at ?p ?to)
        (clear ?from))
)

(:action push-to-goal
    :parameters (?p - thing ?s - thing ?ppos - location ?from - location ?to - location ?dir - direction)
    :precondition (and (move ?dir)
        (is-player ?p)
        (is-stone ?s)
        (at ?p ?ppos)
        (at ?s ?from)
        (clear ?to)
        (move-dir ?ppos ?from ?dir)
        (move-dir ?from ?to ?dir)
        (is-goal ?to))
    :effect (and
        (not (at ?p ?ppos))
        (not (at ?s ?from))
        (not (clear ?to))
        (at ?p ?from)
        (at ?s ?to)
        (clear ?ppos)
        (at-goal ?s))
)


(:action push-to-nongoal
    :parameters (?p - thing ?s - thing ?ppos - location ?from - location ?to - location ?dir - direction)
    :precondition (and (move ?dir)
        (is-player ?p)
        (is-stone ?s)
        (at ?p ?ppos)
        (at ?s ?from)
        (clear ?to)
        (move-dir ?ppos ?from ?dir)
        (move-dir ?from ?to ?dir)
        (is-nongoal ?to))
    :effect (and
        (not (at ?p ?ppos))
        (not (at ?s ?from))
        (not (clear ?to))
        (at ?p ?from)
        (at ?s ?to)
        (clear ?ppos)
        (not (at-goal ?s)))
)
)
       """
        return pddl_str.split("\n")

    def state_goal_to_pddl_inst(self, state: SokobanState, goal: SokobanGoal) -> List[str]:
        model: Model = self.goal_to_model([goal])[0]
        inst_l: List[str] = ["(define(problem sokobaninst)", "(:domain sokoban)"]
        dir_objects: List[str] = ["dir-down - direction", "dir-up - direction", "dir-left - direction",
                                  "dir-right - direction"]
        box_objects: List[str] = [f"stone-{i} - thing" for i in range(self.num_boxes)]
        pos_objects: List[str] = []
        for idx1 in range(self.dim):
            for idx2 in range(self.dim):
                pos_objects.append(f"pos-{idx1}-{idx2} - location")
        inst_l.append("(:objects")
        for obj in dir_objects + pos_objects + ["player-1 - thing "] + box_objects:
            inst_l.append(obj)
        inst_l.append(")")

        inst_l.append("(:init")
        inst_l.append(f"(at player-1 pos-{state.agent[0]}-{state.agent[1]})")
        idxs1, idxs2 = np.where(state.boxes)
        for box_idx, (idx1, idx2) in enumerate(zip(list(idxs1), list(idxs2))):
            inst_l.append(f"(at stone-{box_idx} pos-{idx1}-{idx2})")

        states_np = self.states_goals_to_nnet_input([state], [goal])[0][0]
        states_np_any = states_np.any(axis=0)
        for idx1 in range(self.dim):
            for idx2 in range(self.dim):
                if states_np_any[idx1, idx2] == 0:
                    inst_l.append(f"(clear pos-{idx1}-{idx2})")

        goal_boxes_np: NDArray[np.uint8] = self._models_to_np([model])[0][1]
        for idx1 in range(self.dim):
            for idx2 in range(self.dim):
                if goal_boxes_np[idx1, idx2] == 1:
                    inst_l.append(f"(is-goal pos-{idx1}-{idx2})")
                else:
                    inst_l.append(f"(is-nongoal pos-{idx1}-{idx2})")

        inst_l.append("(is-player player-1)")
        for box_idx in range(self.num_boxes):
            inst_l.append(f"(is-stone stone-{box_idx})")
        inst_l.append("(move dir-down) (move dir-up) (move dir-right) (move dir-left)")

        for action, act_name in zip([0, 1, 2, 3], ["left", "right", "down", "up"]):
            for idx1 in range(self.dim):
                for idx2 in range(self.dim):
                    curr_idxs = np.array([[idx1, idx2]])
                    next_idxs = self._get_next_idx(curr_idxs, [SkAction(action)])[0]
                    if np.all(curr_idxs[0] == next_idxs):
                        continue
                    inst_l.append(f"(move-dir pos-{idx1}-{idx2} pos-{next_idxs[0]}-{next_idxs[1]} dir-{act_name})")
        inst_l.append(")")

        inst_l.append("(:goal")
        inst_l.append("(and")
        for box_idx in range(self.num_boxes):
            inst_l.append(f"(at-goal stone-{box_idx})")
        inst_l.append(")")
        inst_l.append(")")

        inst_l.append(")")
        return inst_l

    def pddl_action_to_action(self, pddl_action: str) -> int:
        str_to_act: Dict[str, int] = {"left": 0, "right": 1, "down": 2, "up": 3}
        re_res = re.search(r".*dir-(\S+).*", pddl_action)
        assert re_res is not None
        act_str: str = re_res.group(1)
        return str_to_act[act_str]

    def get_v_nnet(self) -> HeurFnNNet:
        nnet = NNet(5000, 1000, 4, 1, True, "V")
        return nnet

    def get_q_nnet(self) -> HeurFnNNet:
        nnet = NNet(5000, 1000, 4, 4, True, "Q")
        return nnet

    def start_state_fixed(self, states: List[SokobanState]) -> List[Model]:
        models: List[Model] = self.state_to_model(states)
        fixed_l: List[Model] = []
        for model in models:
            fixed_atoms: List[Atom] = []
            for atom in model:
                if atom[0] == 'wall':
                    fixed_atoms.append(atom)

            fixed_l.append(frozenset(fixed_atoms))
        return fixed_l

    def get_ground_atoms(self) -> List[Atom]:
        ground_atoms: List[Atom] = []
        for pos_i in range(self.dim):
            for pos_j in range(self.dim):
                ground_atoms.append(("agent", f"{pos_i}", f"{pos_j}"))
                ground_atoms.append(("box", f"{pos_i}", f"{pos_j}"))

        return ground_atoms

    def on_model(self, m) -> Model:
        symbs: Set[str] = set(str(x) for x in m.symbols(shown=True))
        atoms: List[Atom] = []
        for symb in symbs:
            symb = misc_utils.remove_all_whitespace(symb)

            match = re.search(r"agent\((\S+),(\S+)\)", symb)
            atom: Atom
            if match is not None:
                atom = ("agent", match.group(1), match.group(2))
                atoms.append(atom)

            match = re.search(r"box\((\S+),(\S+)\)", symb)
            if match is not None:
                atom = ("box", match.group(1), match.group(2))
                atoms.append(atom)

            match = re.search(r"wall\((\S+),(\S+)\)", symb)
            if match is not None:
                atom = ("wall", match.group(1), match.group(2))
                atoms.append(atom)

        model: Model = frozenset(atoms)
        return model

    def get_bk(self) -> List[str]:
        bk: List[str] = ["pos_x(0..9).", "pos_y(0..9).", "pos_min_x(0).", "pos_max_x(9).", "pos_min_y(0).",
                         "pos_max_y(9).", "at_edge_x(X) :- pos_min_x(X).", "at_edge_x(X) :- pos_max_x(X).",
                         "at_edge_y(X) :- pos_min_y(X).", "at_edge_y(X) :- pos_max_y(X).",
                         "validpos2(X,Y,Xr,Yr) :- pos_x(X), pos_y(Y), pos_x(Xr), pos_y(Yr)."]

        bk.extend([
            "dir(up).", "dir(down).", "dir(left).", "dir(right).",
            "manhat(up).", "manhat(down).", "manhat(left).", "manhat(right).",
            "adj(up,right).", "adj(up,left).", "adj(down,right).", "adj(down,left).", "adj(X,Y) :- adj(Y,X).",
            "opp(up,down).", "opp(left,right).", "opp(X,Y) :- opp(Y,X).",

            "rel(X,Y,Xr,Y,up) :- validpos2(X,Y,Xr,Y), Xr=X+1.",
            "rel(X,Y,Xr,Y,down) :- validpos2(X,Y,Xr,Y), Xr=X-1.",
            "rel(X,Y,X,Yr,left) :- validpos2(X,Y,X,Yr), Yr=Y-1.",
            "rel(X,Y,X,Yr,right) :- validpos2(X,Y,X,Yr), Yr=Y+1.",

            "immovable(X,Y) :- wall(X,Y).",
            "immovable(X,Y) :- box_stuck(X,Y).",
            "at_edge_d(X,Y,up) :- pos_max_x(X), pos_y(Y).",
            "at_edge_d(X,Y,down) :- pos_min_x(X), pos_y(Y).",
            "at_edge_d(X,Y,left) :- pos_x(X), pos_min_y(Y).",
            "at_edge_d(X,Y,right) :- pos_x(X), pos_max_y(Y).",
            "immovable_edge_d(X,Y,D) :- rel(X,Y,Xr,Yr,D), immovable(Xr,Yr).",
            "immovable_edge_d(X,Y,D) :- at_edge_d(X,Y,D).",

            "box_stuck(X,Y) :- box(X,Y), manhat(D1), immovable_edge_d(X,Y,D1), adj(D1,D2), "
            "immovable_edge_d(X,Y,D2).",

            ":- box(X,Y), wall(X,Y).",
            ":- box(X,Y), agent(X,Y).",
            ":- wall(X,Y), agent(X,Y).",
        ])
        bk.append("#defined agent/2")
        bk.append("#defined box/2")
        bk.append("#defined wall/2")
        """
        bk.extend([
            "adj(O1,O2) :- at(O1,X1,Y1), at(O2,X2,Y2), adj(X1,Y1,X2,Y2).",
            "box_stuck(B) :- #count{Obj: adj(B, Obj), immovable(Obj)} >= 2.",
        ])
        """

        return bk

    def get_render_array(self, state: Union[SokobanState, SokobanGoal]) -> NDArray[np.int_]:
        state_rendered: NDArray[np.int_] = np.ones((self.dim, self.dim), dtype=int)
        if type(state) is SokobanState:
            state_rendered -= state.walls
            state_rendered[state.agent[0], state.agent[1]] = 2
            state_rendered += state.boxes * 2
        else:
            model: Model = self.goal_to_model([cast(SokobanGoal, state)])[0]
            model_np: NDArray[np.uint8] = self._models_to_np([model])[0]
            state_rendered -= model_np[2]
            state_rendered = state_rendered * (1 - model_np[0]) + 2 * model_np[0]
            state_rendered += model_np[1] * 2

        """
        state_rendered += (state.boxes & state.goals) * 5
        state_rendered += (state.boxes & np.logical_not(state.goals)) * 3

        state_rendered += (state.goals & np.logical_not(state.boxes)) * 1

        if state.goals[state.agent[0], state.agent[1]]:
            state_rendered[state.agent[0], state.agent[1]] = 5
        else:
            state_rendered[state.agent[0], state.agent[1]] = 3
        """

        return state_rendered

    def visualize(self, states: Union[List[SokobanState], List[SokobanGoal]]) -> NDArray[np.float64]:
        states_img: NDArray[np.float64] = np.zeros((len(states), self.img_dim, self.img_dim, 3))

        from PIL import Image
        if self._surfaces is None:
            self._surfaces = _get_surfaces()

        state: Union[SokobanState, SokobanGoal]
        for state_idx, state in enumerate(states):  # type: ignore
            room: NDArray[np.int_] = self.get_render_array(state)

            # Assemble the new rgb_room, with all loaded images
            room_rgb: NDArray[np.uint8] = np.zeros(shape=(room.shape[0] * 16, room.shape[1] * 16, 3), dtype=np.uint8)
            for i in range(room.shape[0]):
                x_i = i * 16

                for j in range(room.shape[1]):
                    y_j = j * 16
                    surfaces_id = room[i, j]

                    room_rgb[x_i:(x_i + 16), y_j:(y_j + 16), :] = self._surfaces[surfaces_id]

            img = Image.fromarray(room_rgb, 'RGB')
            img = img.resize((self.img_dim, self.img_dim))
            states_img[state_idx] = np.array(img) / 255
            # states_img[state_idx] = cv2.resize(room_rgb / 255, (self.img_dim, self.img_dim))

        return states_img

    def _models_to_np(self, models: List[Model]) -> NDArray[np.uint8]:
        goals_rep: NDArray[np.uint8] = np.zeros((len(models), 3, self.dim, self.dim), np.uint8)
        for idx, model in enumerate(models):
            agent_idx = np.array([atom[1:] for atom in model if atom[0] == "agent"]).astype(int)
            box_idxs = np.array([atom[1:] for atom in model if atom[0] == "box"]).astype(int)
            wall_idxs = np.array([atom[1:] for atom in model if atom[0] == "wall"]).astype(int)

            if agent_idx.shape[0] > 0:
                goals_rep[idx, 0, agent_idx[:, 0], agent_idx[:, 1]] = 1
            if box_idxs.shape[0] > 0:
                goals_rep[idx, 1, box_idxs[:, 0], box_idxs[:, 1]] = 1
            if wall_idxs.shape[0] > 0:
                goals_rep[idx, 2, wall_idxs[:, 0], wall_idxs[:, 1]] = 1

        return goals_rep

    def _get_next_idx(self, curr_idxs: NDArray[np.int_], actions: List[SkAction]) -> NDArray[np.int_]:
        actions_np: NDArray[np.int_] = np.array([action.action for action in actions])
        next_idxs: NDArray[np.int_] = curr_idxs.copy()

        action_idxs: NDArray[np.int_] = np.where(actions_np == 0)[0]
        next_idxs[action_idxs, 0] = next_idxs[action_idxs, 0] - 1

        action_idxs = np.where(actions_np == 1)[0]
        next_idxs[action_idxs, 0] = next_idxs[action_idxs, 0] + 1

        action_idxs = np.where(actions_np == 2)[0]
        next_idxs[action_idxs, 1] = next_idxs[action_idxs, 1] - 1

        action_idxs = np.where(actions_np == 3)[0]
        next_idxs[action_idxs, 1] = next_idxs[action_idxs, 1] + 1

        next_idxs = np.maximum(next_idxs, 0)
        next_idxs = np.minimum(next_idxs, self.dim - 1)

        return next_idxs

    def _states_to_np(self, states: List[SokobanState]) -> NDArray[np.uint8]:
        np_rep: NDArray[np.uint8] = np.zeros((len(states), 3, self.dim, self.dim), dtype=np.uint8)
        for idx, state_goal in enumerate(states):
            np_rep[idx, 0, state_goal.agent[0], state_goal.agent[1]] = 1
            np_rep[idx, 1, :, :] = state_goal.boxes
            np_rep[idx, 2, :, :] = state_goal.walls

        return np_rep

    def _goals_to_np(self, goals: List[SokobanGoal]) -> NDArray[np.uint8]:
        np_rep: NDArray[np.uint8] = np.zeros((len(goals), 3, self.dim, self.dim), dtype=np.uint8)
        for idx, state_goal in enumerate(goals):
            if state_goal.agent.shape[0] > 0:
                np_rep[idx, 0, state_goal.agent[0], state_goal.agent[1]] = 1
            np_rep[idx, 1, :, :] = state_goal.boxes
            np_rep[idx, 2, :, :] = state_goal.walls

        return np_rep

    def __getstate__(self):
        self.states_train = None
        self._surfaces = None

        return self.__dict__


class InteractiveEnv(plt.Axes):  # type: ignore
    def __init__(self, env, fig):
        self.env: Sokoban = env

        super(InteractiveEnv, self).__init__(plt.gcf(), [0, 0, 1, 1])

        callbacks = fig.canvas.callbacks.callbacks
        del callbacks['key_press_event']

        self.figure.canvas.mpl_connect('key_press_event', self._key_press)

        self._get_instance()
        self._update_plot()

        self.move = []

    def _get_instance(self):
        states, goals = self.env.get_start_goal_pairs([1000])
        self.state: SokobanState = states[0]
        self.goal: SokobanGoal = goals[0]

    def _update_plot(self):
        self.clear()
        rendered_im = self.env.visualize([self.state])[0]
        rendered_im_goal = self.env.visualize([self.goal])[0]

        self.imshow(np.concatenate((rendered_im, rendered_im_goal), axis=1))
        self.figure.canvas.draw()

    def _key_press(self, event):
        if event.key.upper() in 'ASDW':
            action: int = -1
            if event.key.upper() == 'W':
                action = 0
            if event.key.upper() == 'S':
                action = 1
            if event.key.upper() == 'A':
                action = 2
            if event.key.upper() == 'D':
                action = 3

            self.state = self.env.next_state([self.state], [SkAction(action)])[0][0]
            self._update_plot()
            if self.env.is_solved([self.state], [self.goal])[0]:
                print("SOLVED!")
        elif event.key.upper() in 'R':
            self._get_instance()
            self._update_plot()
        elif event.key.upper() in 'P':
            for i in range(1000):
                self.state = self.env.next_state_rand([self.state])[0][0]
            self._update_plot()


def main():
    env: Sokoban = Sokoban("sokoban")

    fig = plt.figure(figsize=(5, 5))
    interactive_env = InteractiveEnv(env, fig)
    fig.add_axes(interactive_env)

    plt.show()


if __name__ == '__main__':
    main()



from typing import List, Dict, Tuple, Union, Any, Set, Optional
from deepxube.utils import misc_utils
from deepxube.nnet.pytorch_models import FullyConnectedModel, ResnetModel
from deepxube.logic.logic_objects import Atom, Model
from deepxube.visualizers.cube3_viz_simple import InteractiveCube
from .environment_abstract import EnvGrndAtoms, State, Action, Goal, HeurFnNNet

import numpy as np
import torch
from torch import nn, Tensor
import torch.nn.functional as F

from random import randrange

from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
import matplotlib.pyplot as plt
from itertools import permutations
import re
from numpy.typing import NDArray


class Cube3ProcessStates(nn.Module):
    def _forward_unimplemented(self, *input_val: Any) -> None:
        pass

    def __init__(self, state_dim: int, one_hot_depth: int):
        super().__init__()
        self.state_dim: int = state_dim
        self.one_hot_depth: int = one_hot_depth

    def forward(self, states_nnet: Tensor):
        x = states_nnet

        # preprocess input
        if self.one_hot_depth > 0:
            x = F.one_hot(x.long(), self.one_hot_depth)
            x = x.float()
            x = x.view(-1, self.state_dim * self.one_hot_depth)
        else:
            x = x.float()

        return x


class Cube3FCResnet(nn.Module):
    def __init__(self, input_dim: int, h1_dim: int, resnet_dim: int, num_resnet_blocks: int, out_dim: int,
                 batch_norm: bool, weight_norm: bool):
        super().__init__()
        self.first_fc = FullyConnectedModel(input_dim, [h1_dim, resnet_dim], [batch_norm] * 2, ["RELU"] * 2,
                                            weight_norms=[weight_norm] * 2)
        self.resnet = ResnetModel(resnet_dim, num_resnet_blocks, out_dim, batch_norm, weight_norm=weight_norm,
                                  layer_act="RELU")

    def forward(self, x: Tensor):
        x = self.first_fc(x)
        x = self.resnet(x)

        return x


class Cube3NNet(HeurFnNNet):
    def __init__(self, state_dim: int, one_hot_depth: int, h1_dim: int, resnet_dim: int, num_res_blocks: int,
                 out_dim: int, batch_norm: bool, weight_norm: bool, nnet_type: str):
        super().__init__(nnet_type)
        self.state_proc = Cube3ProcessStates(state_dim, one_hot_depth)

        input_dim: int = state_dim * one_hot_depth * 2
        self.heur = Cube3FCResnet(input_dim, h1_dim, resnet_dim, num_res_blocks, out_dim, batch_norm, weight_norm)

    def forward(self, states_goals_l: List[Tensor]):
        states_proc = self.state_proc(states_goals_l[0])
        goals_proc = self.state_proc(states_goals_l[1])

        x: Tensor = self.heur(torch.cat((states_proc, goals_proc), dim=1))

        return x


class Cube3State(State):
    __slots__ = ['colors', 'hash']

    def __init__(self, colors: NDArray[np.uint8]):
        self.colors: NDArray[np.uint8] = colors
        self.hash: Optional[int] = None

    def __hash__(self):
        if self.hash is None:
            self.hash = hash(self.colors.tobytes())
        return self.hash

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Cube3State):
            return np.array_equal(self.colors, other.colors)
        return NotImplemented


class Cube3Goal(Goal):
    def __init__(self, colors: NDArray[np.uint8]):
        self.colors: NDArray[np.uint8] = colors


class Cube3Action(Action):
    def __init__(self, action: int):
        self.action = action

    def __hash__(self):
        return self.action

    def __eq__(self, other: object):
        if isinstance(other, Cube3Action):
            return self.action == other.action
        return NotImplemented


def _get_adj() -> Dict[int, NDArray[np.int_]]:
    # WHITE:0, YELLOW:1, BLUE:2, GREEN:3, ORANGE: 4, RED: 5
    return {0: np.array([2, 5, 3, 4]),
            1: np.array([2, 4, 3, 5]),
            2: np.array([0, 4, 1, 5]),
            3: np.array([0, 5, 1, 4]),
            4: np.array([0, 3, 1, 2]),
            5: np.array([0, 2, 1, 3])
            }


def _cubelet_to_type(cubelet: List[int]) -> str:
    if len(cubelet) == 1:
        return "center_cbl"
    elif len(cubelet) == 2:
        return "edge_cbl"
    elif len(cubelet) == 3:
        return "corner_cbl"
    else:
        raise ValueError("Unknown cubelet type")


def _colors_to_model(colors: NDArray[np.uint8]) -> Model:
    grnd_atoms: List[Atom] = [('at_idx', x, f"{i}") for i, x in enumerate(colors) if x != 'k']
    return frozenset(grnd_atoms)


class Cube3(EnvGrndAtoms[Cube3State, Cube3Action, Cube3Goal]):
    atomic_actions: List[str] = ["%s%i" % (f, n) for f in ['U', 'D', 'L', 'R', 'B', 'F'] for n in [-1, 1]]

    def __init__(self, env_name: str):
        super().__init__(env_name)
        self.cube_len: int = 3
        self.colors: List[str] = ['white', 'yellow', 'orange', 'red', 'blue', 'green']
        self.colors_grnd_obj: List[str] = ['w', 'y', 'o', 'r', 'b', 'g']
        self.faces_grnd_obj: List[str] = ['w_f', 'y_f', 'o_f', 'r_f', 'b_f', 'g_f']

        self.color_to_int: Dict[str, int] = dict()
        for i, col_abbrev in enumerate(self.colors_grnd_obj):
            self.color_to_int[col_abbrev] = i

        self.cbs_m_idxs_l: List[List[int]] = [[4], [13], [22], [31], [40], [49]]
        self.cbs_e_idxs_l: List[List[int]] = [[1, 23], [3, 50], [5, 41], [7, 32],
                                              [10, 21], [12, 39], [14, 48], [16, 30],
                                              [19, 43], [25, 46],
                                              [28, 52], [34, 37]]
        self.cbs_c_idxs_l: List[List[int]] = [[0, 26, 47], [2, 20, 44], [6, 29, 53], [8, 35, 38],
                                              [9, 18, 42], [11, 24, 45], [15, 33, 36], [17, 27, 51]]

        self.cbls_all: List[List[int]] = self.cbs_m_idxs_l.copy()
        self.cbls_all = self.cbs_m_idxs_l + self.cbs_e_idxs_l + self.cbs_c_idxs_l

        # all actions
        # for i in range(0, len(self.atomic_actions), 2):
        #    self.action_combs.append([i, i])

        # self.action_combs = action_combs
        self.num_actions = len(self.atomic_actions)
        self.num_stickers: int = 6 * (self.cube_len ** 2)

        # solved state
        self.goal_colors: NDArray[np.uint8] = (np.arange(0, self.num_stickers, 1,
                                                         dtype=np.uint8) // (self.cube_len ** 2)).astype(np.uint8)

        # get idxs changed for moves
        self.rotate_idxs_new: Dict[str, NDArray[np.int_]]
        self.rotate_idxs_old: Dict[str, NDArray[np.int_]]

        self.adj_faces: Dict[int, NDArray[np.int_]] = _get_adj()

        self.rotate_idxs_new, self.rotate_idxs_old = self._compute_rotation_idxs(self.cube_len, self.atomic_actions)

        self.int_to_color: NDArray[np.str_] = np.concatenate((np.array(self.colors_grnd_obj), ['k']))  # type: ignore

    def next_state(self, states: List[Cube3State], actions: List[Cube3Action]) -> Tuple[List[Cube3State], List[float]]:
        states_np = np.stack([x.colors for x in states], axis=0)

        states_next_np = np.zeros(states_np.shape, dtype=np.uint8)
        tcs_np: NDArray[np.float64] = np.zeros(len(states))
        for action in set(actions):
            action_idxs: NDArray[np.int_] = np.array([idx for idx in range(len(actions)) if actions[idx] == action])
            states_np_act = states_np[action_idxs]

            states_next_np_act = self._move_np(states_np_act, action.action)

            tcs_act: List[float] = [1.0 for _ in range(states_np_act.shape[0])]

            states_next_np[action_idxs] = states_next_np_act
            tcs_np[action_idxs] = np.array(tcs_act)

        states_next: List[Cube3State] = [Cube3State(x) for x in list(states_next_np)]
        transition_costs = list(tcs_np)

        return states_next, transition_costs

    def get_state_actions(self, states: List[Cube3State]) -> List[List[Cube3Action]]:
        return [[Cube3Action(x) for x in range(self.num_actions)] for _ in range(len(states))]

    def is_solved(self, states: List[Cube3State], goals: List[Cube3Goal]) -> List[bool]:
        states_np = np.stack([x.colors for x in states], axis=0)
        goals_np = np.stack([x.colors for x in goals], axis=0)
        is_solved_np = np.all(np.logical_or(states_np == goals_np, goals_np == 6), axis=1)
        return list(is_solved_np)

    def states_goals_to_nnet_input(self, states: List[Cube3State], goals: List[Cube3Goal]) -> List[NDArray[np.uint8]]:
        # states
        states_np: NDArray[np.uint8] = np.stack([state.colors for state in states], axis=0).astype(np.uint8)
        goals_np: NDArray[np.uint8] = np.stack([goal.colors for goal in goals], axis=0)
        return [states_np, goals_np]

    def state_to_model(self, states: List[Cube3State]) -> List[Model]:
        states_np = np.stack([x.colors for x in states], axis=0).astype(np.uint8)
        colors_l = self.int_to_color[states_np]
        models: List[Model] = [_colors_to_model(x) for x in colors_l]

        return models

    def model_to_state(self, states_m: List[Model]) -> List[Cube3State]:
        for state_m in states_m:
            assert len(state_m) == self.num_stickers, "model should be fully specified"
        return [Cube3State(x) for x in self._models_to_np(states_m)]

    def goal_to_model(self, goals: List[Cube3Goal]) -> List[Model]:
        goals_np = np.stack([x.colors for x in goals], axis=0).astype(np.uint8)
        colors_l = self.int_to_color[goals_np]
        models: List[Model] = [_colors_to_model(x) for x in colors_l]

        return models

    def model_to_goal(self, models: List[Model]) -> List[Cube3Goal]:
        return [Cube3Goal(x) for x in self._models_to_np(models)]

    def get_v_nnet(self) -> HeurFnNNet:
        state_dim: int = (self.cube_len ** 2) * 6
        nnet = Cube3NNet(state_dim, 7, 5000, 1000, 4, 1, True, False, "V")

        return nnet

    def get_q_nnet(self) -> HeurFnNNet:
        state_dim: int = (self.cube_len ** 2) * 6
        nnet = Cube3NNet(state_dim, 7, 5000, 1000, 4, self.num_actions, True, False, "Q")

        return nnet

    def get_start_states(self, num_states: int) -> List[Cube3State]:
        assert (num_states > 0)
        backwards_range: Tuple[int, int] = (100, 200)

        # Initialize
        scrambs: List[int] = list(range(backwards_range[0], backwards_range[1] + 1))
        num_atomic_moves: int = len(self.atomic_actions)

        # Get numpy goal states
        goal_np: NDArray[np.uint8] = np.expand_dims(self.goal_colors.copy(), 0)
        states_np: NDArray[np.uint8] = np.repeat(goal_np, num_states, axis=0)

        # Scrambles
        scramble_nums: NDArray[np.int_] = np.random.choice(scrambs, num_states)
        num_back_moves: NDArray[np.int_] = np.zeros(num_states, dtype=int)

        # Go backward from goal state
        moves_lt = num_back_moves < scramble_nums
        while np.any(moves_lt):
            idxs: NDArray[np.int_] = np.where(moves_lt)[0]
            subset_size: int = int(max(len(idxs) / num_atomic_moves, 1))
            idxs = np.random.choice(idxs, subset_size)

            move: int = randrange(num_atomic_moves)
            states_np[idxs] = self._move_np(states_np[idxs], move)

            num_back_moves[idxs] = num_back_moves[idxs] + 1
            moves_lt[idxs] = num_back_moves[idxs] < scramble_nums[idxs]

        states: List[Cube3State] = [Cube3State(x) for x in list(states_np)]

        return states

    def start_state_fixed(self, states: List[Cube3State]) -> List[Model]:
        return [frozenset() for _ in states]

    def get_bk(self) -> List[str]:
        bk: List[str] = []

        clockwise: Dict[str, List[str]] = dict()
        clockwise["w"] = ["b", "r", "g", "o"]
        clockwise["o"] = ["y", "b", "w", "g"]
        clockwise["g"] = ["o", "w", "r", "y"]

        opposite: Dict[str, str] = {"w": "y", "o": "r", "g": "b"}

        bk.append("% Colors")
        for color_obj in self.colors_grnd_obj:
            bk.append(f"color({color_obj})")
        bk.append("dif_col(Col1, Col2) :- color(Col1), color(Col2), Col1 != Col2")

        bk.append("\n")
        for color_idx, (color, color_obj) in enumerate(zip(self.colors, self.colors_grnd_obj)):
            bk.append(f"{color}({color_obj})")

        bk.append("\n% Cubelets")

        center_cbls = [x for x in self.cbls_all if len(x) == 1]
        for cbl_l in center_cbls:
            bk.append(f"center_cbl({self._cubelet_to_name(cbl_l)})")

        edge_cbls = [x for x in self.cbls_all if len(x) == 2]
        for cbl_l in edge_cbls:
            bk.append(f"edge_cbl({self._cubelet_to_name(cbl_l)})")

        corner_cbls = [x for x in self.cbls_all if len(x) == 3]
        for cbl_l in corner_cbls:
            bk.append(f"corner_cbl({self._cubelet_to_name(cbl_l)})")

        if len(center_cbls) > 0:
            bk.append("cubelet(Cbl) :- center_cbl(Cbl)")
        if len(edge_cbls) > 0:
            bk.append("cubelet(Cbl) :- edge_cbl(Cbl)")
        if len(corner_cbls) > 0:
            bk.append("cubelet(Cbl) :- corner_cbl(Cbl)")
        bk.append("dif_cbl(Cbl1, Cbl2) :- cubelet(Cbl1), cubelet(Cbl2), Cbl1 != Cbl2")

        bk.append("\n% has_stk_col\n")
        for cbl_l in self.cbls_all:
            cbl_name: str = self._cubelet_to_name(cbl_l)
            color_objs: List[str] = self._cubelet_to_colors_obj(cbl_l)
            for color_obj in color_objs:
                bk.append(f"has_stk_col({cbl_name}, {color_obj})")

        """
        for cbl_idxs in self.cbs_m_idxs_l + self.cbs_e_idxs_l + self.cbs_c_idxs_l:
            face_names: List[str] = [f"{self.colors_abbrev[idx // (self.cube_len ** 2)]}_f" for idx in cbl_idxs]
            for idx, face_name in zip(cbl_idxs, face_names):
                bk.append(f"onface(Cbl, Col, {face_name}) :- at_idx_cbl(Cbl, Col, {idx})")

        for cbl_idxs in self.cbs_m_idxs_l + self.cbs_e_idxs_l + self.cbs_c_idxs_l:
            for idx_1 in range(len(cbl_idxs)):
                for idx_2 in range(idx_1 + 1, len(cbl_idxs)):
                    bk.append(f":- at_idx_cbl(Cbl0, _, {cbl_idxs[idx_1]}), at_idx_cbl(Cbl1, _, {cbl_idxs[idx_2]}), "
                              f"dif_cbl(Cbl0, Cbl1)")

        for cbl_idxs in self.cbs_m_idxs_l + self.cbs_e_idxs_l + self.cbs_c_idxs_l:
            if len(cbl_idxs) == 1:
                continue
            neq_str: str = ", ".join([f"I != {cbl_idx}" for cbl_idx in cbl_idxs])
            for cbl_idx in cbl_idxs:
                bk.append(f":- at_idx_cbl(Cbl, _, {cbl_idx}), at_idx_cbl(Cbl, _, I), {neq_str}")
        """

        bk.append("\n% Directions")
        for direction in ["cl", "cc", "op"]:
            bk.append(f"direction({direction})")
        bk.append("dif_dir(Dir1, Dir2) :- direction(Dir1), direction(Dir2), Dir1 != Dir2")

        bk.append("")
        bk.extend(["clockwise(cl)", "counterclockwise(cc)", "opposite(op)"])

        bk.append("\n% Faces")
        for color_obj in self.colors_grnd_obj:
            bk.append(f"face({color_obj}_f)")
        bk.append("dif_face(F1, F2) :- face(F1), face(F2), F1 != F2")

        bk.append("\n")
        for color_obj in self.colors_grnd_obj:
            bk.append(f"face_col({color_obj}_f, {color_obj})")

        bk.append("\n")
        for center in clockwise.keys():
            centers_cl: List[str] = clockwise[center]
            for idx, center_cl in enumerate(centers_cl):
                idx_next = (idx + 1) % len(centers_cl)
                bk.append(f"face_rel({center}_f, {center_cl}_f, {centers_cl[idx_next]}_f, cl)")

        for center_op in clockwise.keys():
            center = opposite[center_op]
            centers_cl = clockwise[center_op]
            for idx, center_cl in enumerate(centers_cl):
                idx_next = (idx - 1) % len(centers_cl)
                bk.append(f"face_rel({center}_f, {center_cl}_f, {centers_cl[idx_next]}_f, cl)")
        bk.append("\n")

        bk.append("face_rel(A, B, C, cc) :- face_rel(A, C, B, cl)")
        bk.append("face_rel(A, B, D, op) :- face_rel(A, B, C, cl), face_rel(A, C, D, cl)")
        bk.append("face_rel(A, D, B, op) :- face_rel(A, B, C, cl), face_rel(A, C, D, cl)")
        bk.append("face_adj(F0, F1) :- face_rel(F0, _, F1, _)")

        bk.append("\n% State dependent predicates")
        bk.append("index(0..53)")
        bk.append("-at_idx(Col, I) :- at_idx(Col2, I), color(Col), color(Col2), not Col=Col2")

        bk.append("\n% onface\n")
        for cbl_idxs in self.cbs_m_idxs_l + self.cbs_e_idxs_l + self.cbs_c_idxs_l:
            face_names: List[str] = [f"{self.colors_grnd_obj[idx // (self.cube_len ** 2)]}_f" for idx in cbl_idxs]
            for cbl_idxs2 in self.cbls_all:
                if _cubelet_to_type(cbl_idxs) != _cubelet_to_type(cbl_idxs2):
                    continue
                color_obj_perms = permutations(self._cubelet_to_colors_obj(cbl_idxs2))
                for color_objs_p in color_obj_perms:
                    at_idx_str: str = ", ".join([f"at_idx({col}, {idx})" for col, idx in zip(color_objs_p, cbl_idxs)])
                    for col, face_name in zip(color_objs_p, face_names):
                        bk.append(f"onface({self._cubelet_to_name(cbl_idxs2)}, {col}, {face_name}) :- {at_idx_str}")

        bk.append("stk_f_dir(Cbl, StkCol, FTo, Dir) :- edge_cbl(Cbl), face_rel(FRef, FFrom, FTo, Dir), "
                  "onface(Cbl, StkCol, FFrom), onface(Cbl, _, FRef)")

        bk.append("\n% in_place")
        bk.append("in_place(Cbl) :- edge_cbl(Cbl), onface(Cbl, Col0, F0), face_col(F0, Col0), "
                  "onface(Cbl, Col1, F1), face_col(F1, Col1), Col0 != Col1")
        bk.append("in_place(Cbl) :- corner_cbl(Cbl), onface(Cbl, Col0, F0), face_col(F0, Col0), "
                  "onface(Cbl, Col1, F1), face_col(F1, Col1), onface(Cbl, Col2, F2), face_col(F2, Col2), "
                  "Col0 != Col1, Col0 != Col2, Col1 != Col2")
        bk.append("in_place(Cbl) :- center_cbl(Cbl), onface(Cbl, Col0, F0), face_col(F0, Col0)")

        # bk.append("1 { onface(Cbl, Col1, F1) : face_adj(F0, F1), dif_col(Col0, Col1) } 1 :- "
        #           "edge_cbl(Cbl), onface(Cbl, Col0, F0).\n")
        # bk.append(":- corner_cbl(Cbl), onface(Cbl, Col0, F0), "
        #           "not 2 { onface(Cbl, Col1, F1) : face_adj(F0, F1), dif_col(Col0, Col1) } 2.\n")

        """
        for action in range(self.num_actions):
            bk.append(f"\n% Action {action}")
            # TODO hacky, 0-53 when usually 0-5
            state: Cube3State = Cube3State(np.arange(0, self.num_stickers, 1, dtype=self.dtype))
            state_next: Cube3State = self.next_state([state], [action])[0][0]
            if action < 12:
                act_face: str = self.faces_grnd_obj[action // 2]
                if action % 2 == 0:
                    act_dir: str = "cc"
                else:
                    act_dir: str = "cl"
            else:
                act_face: str = self.faces_grnd_obj[action - 12]
                act_dir: str = "op"
            act_lit_str: str = f"act(T, {act_face}, {act_dir})"

            for color_unique in state.colors:
                idx1: int = np.where(state.colors == color_unique)[0][0]
                idx2: int = np.where(state_next.colors == color_unique)[0][0]
                bk.append(f"at_idx(T+1, Col, {idx2}) :- time(T), at_idx(T, Col, {idx1}), {act_lit_str}")
        """

        bk.append("\n% constraints")
        """
        bk.append("% all center cubelets are always on their respective faces\n")
        for cbl_l in [x for x in self.subgoal_cbls if len(x) == 1]:
            cbl_name: str = self._cubelet_to_name(cbl_l)
            color_obj: str = self._cubelet_to_colors_obj(cbl_l)[0]
            bk.append(f":- not onface({cbl_name}, {color_obj}, {color_obj}_f)")
        """

        bk.append("% different stickers from the same cubelet cannot be on the same face")
        bk.append(":- onface(Cbl, Col0, F), onface(Cbl, Col1, F), dif_col(Col0, Col1)")

        bk.append("\n% cannot have a sticker color from same cubelet be on more than one face")
        bk.append(":- onface(Cbl, Col, F0), onface(Cbl, Col, F1), dif_face(F0, F1)")

        bk.append("\n% cannot say a color of a cubelet is on a face if it does not have the color")
        bk.append(":- onface(Cbl, Col, _), not has_stk_col(Cbl, Col)")

        bk.append("\n% cannot have two stickers from same cbl on opposite faces")
        bk.append(":- onface(Cbl, Col0, F0), onface(Cbl, Col1, F1), face_rel(_, F0, F1, op)")

        bk.append("\n% cannot have two different colors at the same index")
        bk.append(":- at_idx(Col0, I), at_idx(Col1, I), dif_col(Col0, Col1)")

        # bk.append("\n% cannot have two different cubelets at the same index")
        # bk.append(":- at_idx_cbl(Cbl0, _, I), at_idx_cbl(Cbl1, _, I), dif_cbl(Cbl0, Cbl1)")

        """
        bk.append("\n% cannot have stickers from different cubelets on same cubelet loc")
        for cbl in self.subgoal_cbls:
            for cbl_i in cbl:
                for cbl_i2 in cbl:
                    bk.append(f":- at_idx_cbl(Cbl, _, {cbl_i}), at_idx_cbl(CblOth, _, {cbl_i2}), "
                               f"dif_cbl(Cbl, CblOth)\n")

        bk.append("\n% cannot have stickers from same cubelet on different cubelet locs\n")
        for cbl in self.subgoal_cbls:
            for cbl_i in cbl:
                cbl_oths = [x for x in cbl if x != cbl_i]
                if len(cbl) == 2:
                    bk.append(f":- at_idx_cbl(Cbl, _, {cbl_i}), at_idx_cbl(Cbl, _, I), not I={cbl_i}, "
                               f"not I={cbl_oths[0]}\n")
                if len(cbl) == 3:
                    bk.append(f":- at_idx_cbl(Cbl, _, {cbl_i}), at_idx_cbl(Cbl, _, I), not I={cbl_i}, "
                               f"not I={cbl_oths[0]}, not I={cbl_oths[1]}\n")
        """

        # bk.append("count_at_idx(C) :- #count{V0, V1: at_idx(V0,V1) }=C")
        # bk.append("#minimize {C: count_at_idx(C)}")

        bk.append("#show at_idx/2")
        bk.append("#defined at_idx/2")
        bk.append("#defined center_cbl/1")
        bk.append("#defined edge_cbl/1")
        bk.append("#defined corner_cbl/1")

        return bk

    def get_ground_atoms(self) -> List[Atom]:
        ground_atoms: List[Atom] = []
        for color in self.colors_grnd_obj:
            for idx in range(self.num_stickers):
                ground_atoms.append(("at_idx", f"{color}", f"{idx}"))

        # return ["{ at_idx(Col, I) : color(Col), index(I) } 54"]
        return ground_atoms

    def on_model(self, m) -> Model:
        symbs_set: Set[str] = set(str(x) for x in m.symbols(shown=True))
        symbs: List[str] = [misc_utils.remove_all_whitespace(symb) for symb in symbs_set]

        # get atoms
        atoms: List[Atom] = []
        for symb in symbs:
            match = re.search(r"^at_idx\((\S+),(\S+)\)$", symb)
            if match is None:
                continue
            atom: Atom = ("at_idx", match.group(1), match.group(2))
            atoms.append(atom)

        model: Model = frozenset(atoms)
        return model

    def get_pddl_domain(self) -> List[str]:
        pddl_str: str = """
(define (domain cube3)
(:requirements :strips)
(:predicates (at ?c ?i) )"""

        for action in range(self.num_actions):
            action_str_l: List[str] = [f"(:action a{action}", ":effect", "(and"]
            action_name: str = self.atomic_actions[action]
            for idx_old, idx_new in zip(self.rotate_idxs_old[action_name], self.rotate_idxs_new[action_name]):
                action_str_l.append(f"(forall (?x) (when (at ?x i{idx_old}) "
                                    f"(and (not (at ?x i{idx_old})) (at ?x i{idx_new}))))")
                # action_str_l.append(":parameters (?c)")
                # action_str_l.append(f":precondition (at ?c i{idx_old})")
                # action_str_l.append(f":effect (and (not (at ?c i{idx_old})) (at ?c i{idx_new}))")
            action_str_l.append(")")
            action_str_l.append(")")
            action_str: str = '\n'.join(action_str_l)
            pddl_str = f"{pddl_str}\n{action_str}"
        # pddl_str = pddl_str + ("\n(:action A\n"
        #                       ":parameters (?x ?y)\n"
        #                       ":precondition (at ?x ?y)\n"
        #                       "\n:effect (at ?y ?x))")

        pddl_str = pddl_str + "\n)"

        return pddl_str.split("\n")

    def state_goal_to_pddl_inst(self, state: Cube3State, goal: Cube3Goal) -> List[str]:
        model: Model = self.goal_to_model([goal])[0]
        inst_l: List[str] = ["(define(problem cube3inst)", "(:domain cube3)"]
        idx_objects: str = ' '.join([f"i{idx}" for idx in range(self.num_stickers)])
        objects_str: str = f"(:objects {' '.join(self.colors_grnd_obj)} {idx_objects})"
        inst_l.append(objects_str)

        inst_l.append("(:init")
        inst_l.append(' '.join([f'(at {self.int_to_color[color]} i{idx})' for idx, color in enumerate(state.colors)]))
        inst_l.append(")")

        inst_l.append("(:goal")
        inst_l.append("(and")
        inst_l.append(' '.join([f'(at {self.int_to_color[color]} i{idx})' for idx, color
                                in enumerate(self._models_to_np([model])[0]) if color < 6]))
        inst_l.append(")")
        inst_l.append(")")

        inst_l.append(")")
        return inst_l

    def pddl_action_to_action(self, pddl_action: str) -> int:
        match = re.match(r"^a(\d+).*", pddl_action)
        assert match is not None
        return int(match.group(1))

    def visualize(self, states: Union[List[Cube3State], List[Cube3Goal]]) -> NDArray[np.float64]:
        # initialize
        fig = plt.figure(figsize=(.64, .64))
        viz = InteractiveCube(3, self.get_start_states(1)[0].colors)

        fig.add_axes(viz)
        canvas = FigureCanvas(fig)
        width, height = fig.get_size_inches() * fig.get_dpi()
        width = int(width)
        height = int(height)

        states_img: NDArray[np.float64] = np.zeros((len(states), width, height, 6))
        for state_idx, state in enumerate(states):
            # create image
            if isinstance(state, Cube3State):
                viz.new_state(state.colors)
            elif isinstance(state, Cube3Goal):
                model: Model = self.goal_to_model([state])[0]
                viz.new_state(self._models_to_np([model])[0])
            else:
                raise ValueError(f"Unknown input type {type(state)}")

            viz.set_rot(0)
            canvas.draw()
            image1 = np.frombuffer(canvas.tostring_rgb(), dtype='uint8').reshape((width, height, 3)) / 255

            viz.set_rot(1)
            canvas.draw()
            image2 = np.frombuffer(canvas.tostring_rgb(), dtype='uint8').reshape((width, height, 3)) / 255

            states_img[state_idx] = np.concatenate((image1, image2), axis=2)

        plt.close(fig)

        return states_img

    def _models_to_np(self, models: List[Model]) -> NDArray[np.uint8]:
        models_np: NDArray[np.uint8] = (np.ones((len(models), self.num_stickers), dtype=np.uint8) * 6).astype(np.uint8)
        for idx, model in enumerate(models):
            for grnd_atom in model:
                models_np[idx, int(grnd_atom[2])] = self.color_to_int[grnd_atom[1]]

        return models_np

    def _cubelet_to_name(self, cubelet: List[int]) -> str:
        face_idxs = np.floor(np.sort(cubelet) // (self.cube_len ** 2)).astype(int)
        cubelet_name = "".join([self.colors_grnd_obj[idx] for idx in face_idxs])
        cubelet_name = cubelet_name + "_c"

        return cubelet_name

    def _cubelet_to_colors_obj(self, cubelet: List[int]) -> List[str]:
        face_idxs = np.floor(np.sort(cubelet) // (self.cube_len ** 2)).astype(int)
        return [self.colors_grnd_obj[idx] for idx in face_idxs]

    def _move_np(self, states_np: NDArray[np.uint8], action: int) -> NDArray[np.uint8]:
        states_next_np: NDArray[np.uint8] = states_np.copy()

        actions = [action]

        for action_part in actions:
            action_str: str = self.atomic_actions[action_part]
            states_next_np[:, self.rotate_idxs_new[action_str]] = states_next_np[:, self.rotate_idxs_old[action_str]]

        return states_next_np

    def _compute_rotation_idxs(self, cube_len: int,
                               moves: List[str]) -> Tuple[Dict[str, NDArray[np.int_]], Dict[str, NDArray[np.int_]]]:
        rotate_idxs_new: Dict[str, NDArray[np.int_]] = dict()
        rotate_idxs_old: Dict[str, NDArray[np.int_]] = dict()

        for move in moves:
            f: str = move[0]
            sign: int = int(move[1:])

            rotate_idxs_new[move] = np.array([], dtype=int)
            rotate_idxs_old[move] = np.array([], dtype=int)

            colors = np.zeros((6, cube_len, cube_len), dtype=np.int64)
            colors_new = np.copy(colors)

            # WHITE:0, YELLOW:1, BLUE:2, GREEN:3, ORANGE: 4, RED: 5

            adj_idxs = {0: {2: [range(0, cube_len), cube_len - 1], 3: [range(0, cube_len), cube_len - 1],
                            4: [range(0, cube_len), cube_len - 1], 5: [range(0, cube_len), cube_len - 1]},
                        1: {2: [range(0, cube_len), 0], 3: [range(0, cube_len), 0], 4: [range(0, cube_len), 0],
                            5: [range(0, cube_len), 0]},
                        2: {0: [0, range(0, cube_len)], 1: [0, range(0, cube_len)],
                            4: [cube_len - 1, range(cube_len - 1, -1, -1)], 5: [0, range(0, cube_len)]},
                        3: {0: [cube_len - 1, range(0, cube_len)], 1: [cube_len - 1, range(0, cube_len)],
                            4: [0, range(cube_len - 1, -1, -1)], 5: [cube_len - 1, range(0, cube_len)]},
                        4: {0: [range(0, cube_len), cube_len - 1], 1: [range(cube_len - 1, -1, -1), 0],
                            2: [0, range(0, cube_len)], 3: [cube_len - 1, range(cube_len - 1, -1, -1)]},
                        5: {0: [range(0, cube_len), 0], 1: [range(cube_len - 1, -1, -1), cube_len - 1],
                            2: [cube_len - 1, range(0, cube_len)], 3: [0, range(cube_len - 1, -1, -1)]}
                        }
            face_dict = {'U': 0, 'D': 1, 'L': 2, 'R': 3, 'B': 4, 'F': 5}
            face = face_dict[f]

            faces_to = self.adj_faces[face]
            if sign == 1:
                faces_from = faces_to[(np.arange(0, len(faces_to)) + 1) % len(faces_to)]
            else:
                faces_from = faces_to[(np.arange(len(faces_to) - 1, len(faces_to) - 1 + len(faces_to))) % len(faces_to)]

            cubes_idxs = [[0, range(0, cube_len)], [range(0, cube_len), cube_len - 1],
                          [cube_len - 1, range(cube_len - 1, -1, -1)], [range(cube_len - 1, -1, -1), 0]]
            cubes_to = np.array([0, 1, 2, 3])
            if sign == 1:
                cubes_from = cubes_to[(np.arange(len(cubes_to) - 1, len(cubes_to) - 1 + len(cubes_to))) % len(cubes_to)]
            else:
                cubes_from = cubes_to[(np.arange(0, len(cubes_to)) + 1) % len(cubes_to)]

            for i in range(4):
                idxs_new = [[idx1, idx2] for idx1 in np.array([cubes_idxs[cubes_to[i]][0]]).flatten() for idx2 in
                            np.array([cubes_idxs[cubes_to[i]][1]]).flatten()]
                idxs_old = [[idx1, idx2] for idx1 in np.array([cubes_idxs[cubes_from[i]][0]]).flatten() for idx2 in
                            np.array([cubes_idxs[cubes_from[i]][1]]).flatten()]
                for idxNew, idxOld in zip(idxs_new, idxs_old):
                    flat_idx_new: int = int(np.ravel_multi_index((face, idxNew[0], idxNew[1]), colors_new.shape))
                    flat_idx_old: int = int(np.ravel_multi_index((face, idxOld[0], idxOld[1]), colors.shape))
                    rotate_idxs_new[move] = np.concatenate((rotate_idxs_new[move], [flat_idx_new]))  # type: ignore
                    rotate_idxs_old[move] = np.concatenate((rotate_idxs_old[move], [flat_idx_old]))  # type: ignore

            # Rotate adjacent faces
            face_idxs = adj_idxs[face]
            for i in range(0, len(faces_to)):
                face_to: int = int(faces_to[i])
                face_from: int = int(faces_from[i])
                idxs_new = [[idx1, idx2] for idx1 in np.array([face_idxs[face_to][0]]).flatten() for idx2 in
                            np.array([face_idxs[face_to][1]]).flatten()]
                idxs_old = [[idx1, idx2] for idx1 in np.array([face_idxs[face_from][0]]).flatten() for idx2 in
                            np.array([face_idxs[face_from][1]]).flatten()]
                for idxNew, idxOld in zip(idxs_new, idxs_old):
                    flat_idx_new = int(np.ravel_multi_index((face_to, idxNew[0], idxNew[1]), colors_new.shape))
                    flat_idx_old = int(np.ravel_multi_index((face_from, idxOld[0], idxOld[1]), colors.shape))
                    rotate_idxs_new[move] = np.concatenate((rotate_idxs_new[move], [flat_idx_new]))  # type: ignore
                    rotate_idxs_old[move] = np.concatenate((rotate_idxs_old[move], [flat_idx_old]))  # type: ignore

        return rotate_idxs_new, rotate_idxs_old



# ----------------------------------------------------------------------
# Matplotlib Rubik's cube simulator
# Written by Jake Vanderplas
# Adapted from cube code written by David Hogg
#   https://github.com/davidwhogg/MagicCube

from typing import List, Optional
import numpy as np
from numpy.typing import NDArray

import matplotlib.pyplot as plt

from deepxube.utils.viz_utils import Quaternion, project_points


class InteractiveCube(plt.Axes):  # type: ignore
    # Define some attributes
    base_face = np.array([[1, 1, 1],
                          [1, -1, 1],
                          [-1, -1, 1],
                          [-1, 1, 1],
                          [1, 1, 1]], dtype=float)
    stickerwidth = 0.9
    stickermargin = 0.5 * (1. - stickerwidth)
    stickerthickness = 0.001
    (d1, d2, d3) = (1 - stickermargin,
                    1 - 2 * stickermargin,
                    1 + stickerthickness)
    base_sticker = np.array([[d1, d2, d3], [d2, d1, d3],
                             [-d2, d1, d3], [-d1, d2, d3],
                             [-d1, -d2, d3], [-d2, -d1, d3],
                             [d2, -d1, d3], [d1, -d2, d3],
                             [d1, d2, d3]], dtype=float)

    base_face_centroid = np.array([[0, 0, 1]])
    base_sticker_centroid = np.array([[0, 0, 1 + stickerthickness]])

    def __init__(self, n, colors: NDArray, view=(0, 0, 10),
                 fig=None, **kwargs):
        self.colors: NDArray = colors

        # Define rotation angles and axes for the six sides of the cube
        x, y, z = np.eye(3)
        self.rots = [Quaternion.from_v_theta(x, theta) for theta in (np.pi / 2, -np.pi / 2)]
        self.rots += [Quaternion.from_v_theta(y, theta) for theta in (np.pi / 2, -np.pi / 2, np.pi, 2 * np.pi)]

        rect = [0, 0.16, 1, 0.84]
        self._move_list: List = []

        self.N = n
        self._prevStates: List = []

        self._view = view
        self._start_rot = Quaternion.from_v_theta((1, -1, 0), -np.pi / 6)

        self._grey_stickers: List = []
        self._black_stickers: List = []

        if fig is None:
            fig = plt.gcf()

        # disable default key press events
        callbacks = fig.canvas.callbacks.callbacks
        del callbacks['key_press_event']

        # add some defaults, and draw axes
        kwargs.update(dict(aspect=kwargs.get('aspect', 'equal'),
                           xlim=kwargs.get('xlim', (-1.7, 1.5)),
                           ylim=kwargs.get('ylim', (-1.5, 1.7)),
                           frameon=kwargs.get('frameon', False),
                           xticks=kwargs.get('xticks', []),
                           yticks=kwargs.get('yticks', [])))
        super(InteractiveCube, self).__init__(fig, rect, **kwargs)
        self.xaxis.set_major_formatter(plt.NullFormatter())  # type: ignore
        self.yaxis.set_major_formatter(plt.NullFormatter())  # type: ignore

        self._start_xlim = kwargs['xlim']
        self._start_ylim = kwargs['ylim']

        # Define movement for up/down arrows or up/down mouse movement
        self._ax_UD = (1, 0, 0)
        self._step_UD = 0.01

        # Define movement for left/right arrows or left/right mouse movement
        self._ax_LR = (0, -1, 0)
        self._step_LR = 0.01

        self._ax_LR_alt = (0, 0, 1)

        self._current_rot = self._start_rot  # current rotation state
        self._face_polys: Optional[List] = None
        self._sticker_polys: Optional[List] = None

        self.plastic_color = 'black'

        # WHITE:0 - U, YELLOW:1 - D, BLUE:2 - L, GREEN:3 - R, ORANGE: 4 - B, RED: 5 - F
        self.face_colors = ["w", "#ffcf00",
                            "#ff6f00", "#cf0000",
                            "#00008f", "#009f0f",
                            "gray", "none"]

        self._initialize_arrays()

        self._draw_cube()
        # self._initialize_widgets()

    def set_rot(self, rot: int):
        if rot == 0:
            self._current_rot = Quaternion.from_v_theta((-0.53180525,  0.83020462,  0.16716299), 0.95063829)
        elif rot == 1:
            self._current_rot = Quaternion.from_v_theta((0.9248325,  0.14011997, -0.35362584), 2.49351394)

        self._draw_cube()

    def _initialize_arrays(self):
        # initialize centroids, faces, and stickers.  We start with a
        # base for each one, and then translate & rotate them into position.

        # Define N^2 translations for each face of the cube
        cubie_width = 2. / self.N
        translations = np.array([[[-1 + (i + 0.5) * cubie_width,
                                   -1 + (j + 0.5) * cubie_width, 0]]
                                 for i in range(self.N)
                                 for j in range(self.N)])

        # Create arrays for centroids, faces, stickers
        face_centroids = []
        faces = []
        sticker_centroids = []
        stickers = []
        colors = []

        factor = np.array([1. / self.N, 1. / self.N, 1])

        for i in range(6):
            rot_mat = self.rots[i].as_rotation_matrix()
            faces_t = np.dot(factor * self.base_face
                             + translations, rot_mat.T)
            stickers_t = np.dot(factor * self.base_sticker
                                + translations, rot_mat.T)
            face_centroids_t = np.dot(self.base_face_centroid
                                      + translations, rot_mat.T)
            sticker_centroids_t = np.dot(self.base_sticker_centroid
                                         + translations, rot_mat.T)
            # colors_i = i + np.zeros(face_centroids_t.shape[0], dtype=int)
            colors_i = np.arange(i * face_centroids_t.shape[0], (i + 1) * face_centroids_t.shape[0])

            # append face ID to the face centroids for lex-sorting
            face_centroids_t = np.hstack([face_centroids_t.reshape(-1, 3),
                                          colors_i[:, None]])
            sticker_centroids_t = sticker_centroids_t.reshape((-1, 3))

            faces.append(faces_t)
            face_centroids.append(face_centroids_t)
            stickers.append(stickers_t)
            sticker_centroids.append(sticker_centroids_t)

            colors.append(colors_i)

        self._face_centroids = np.vstack(face_centroids)
        self._faces = np.vstack(faces)
        self._sticker_centroids = np.vstack(sticker_centroids)
        self._stickers = np.vstack(stickers)

    def _project(self, pts):
        return project_points(pts, self._current_rot, self._view, [0, 1, 0])

    def _draw_cube(self):
        stickers = self._project(self._stickers)[:, :, :2]
        faces = self._project(self._faces)[:, :, :2]
        face_centroids = self._project(self._face_centroids[:, :3])
        sticker_centroids = self._project(self._sticker_centroids[:, :3])

        plastic_color = self.plastic_color
        # self._colors[np.ravel_multi_index((0,1,2),(6,N,N))] = 10
        colors = np.asarray(self.face_colors)[self.colors]
        for idx in self._grey_stickers:
            colors[idx] = "grey"
        for idx in self._black_stickers:
            colors[idx] = "k"

        face_zorders = -face_centroids[:, 2]
        sticker_zorders = -sticker_centroids[:, 2]

        if self._face_polys is None:
            # initial call: create polygon objects and add to axes
            self._face_polys = []
            self._sticker_polys = []

            for i in range(len(colors)):
                fp = plt.Polygon(faces[i], facecolor=plastic_color, zorder=face_zorders[i])  # type: ignore
                sp = plt.Polygon(stickers[i], facecolor=colors[i], zorder=sticker_zorders[i])  # type: ignore

                self._face_polys.append(fp)
                self._sticker_polys.append(sp)
                self.add_patch(fp)
                self.add_patch(sp)
        else:
            assert self._sticker_polys is not None
            # subsequent call: update the polygon objects
            for i in range(len(colors)):
                self._face_polys[i].set_xy(faces[i])
                self._face_polys[i].set_zorder(face_zorders[i])
                self._face_polys[i].set_facecolor(plastic_color)

                self._sticker_polys[i].set_xy(stickers[i])
                self._sticker_polys[i].set_zorder(sticker_zorders[i])
                self._sticker_polys[i].set_facecolor(colors[i])

        self.figure.canvas.draw()

    def new_state(self, colors: NDArray):
        self.colors = colors
        self._draw_cube()



from abc import ABC, abstractmethod
from typing import List, Tuple, Union, Optional, Set, Any, TypeVar, Generic
import numpy as np
import torch.nn as nn
from deepxube.logic.logic_objects import Atom, Model
from deepxube.utils import misc_utils
from deepxube.utils.timing_utils import Times
import random
import time
from numpy.typing import NDArray


class State(ABC):
    @abstractmethod
    def __hash__(self):
        """ For use in CLOSED dictionary for heuristic search
        @return: hash value
        """
        pass

    @abstractmethod
    def __eq__(self, other: object):
        """ for use in state reidentification during heuristic search

        @param other: other state
        @return: true if they are equal
        """
        pass


class Action(ABC):
    pass


class Goal(ABC):
    pass


class HeurFnNNet(nn.Module):
    valid_nnet_types: Set[str] = {"V", "Q"}

    def __init__(self, nnet_type: str):
        """
        @param nnet_type:  "V" for value function for approximate value iteration and A* search.
        "Q" for Q-network for Q-learning and Q* search.
        """
        super().__init__()
        nnet_type = nnet_type.upper()
        assert nnet_type in self.valid_nnet_types, (f"Invalid nnet type: {nnet_type}. Valid nnet types are: "
                                                    f"{self.valid_nnet_types}")
        self.nnet_type = nnet_type


S = TypeVar('S', bound=State)
A = TypeVar('A', bound=Action)
G = TypeVar('G', bound=Goal)


class Environment(ABC, Generic[S, A, G]):
    def __init__(self, env_name: str):
        self.env_name: Optional[str] = env_name

    @abstractmethod
    def get_start_states(self, num_states: int) -> List[S]:
        """ A method for generating start states. Should try to make this generate states that are as diverse as
        possible so that the trained heuristic function generalizes well.

        @param num_states: Number of states to get
        @return: Generated states
        """
        pass

    @abstractmethod
    def get_state_actions(self, states: List[S]) -> List[List[A]]:
        """ Get actions applicable to each states

        @param states: List of states
        @return: Applicable ations
        """
        pass

    @abstractmethod
    def next_state(self, states: List[S], actions: List[A]) -> Tuple[List[S], List[float]]:
        """ Get the next state and transition cost given the current state and action

        @param states: List of states
        @param actions: List of actions to take
        @return: Next states, transition costs
        """
        pass

    def next_state_rand(self, states: List[S]) -> Tuple[List[S], List[float]]:
        """ Get random next state and transition cost given the current state

        @param states: List of states
        @return: Next states, transition costs
        """
        state_actions: List[List[A]] = self.get_state_actions(states)
        actions_rand: List[A] = [random.choice(x) for x in state_actions]
        return self.next_state(states, actions_rand)

    @abstractmethod
    def sample_goal(self, states_start: List[S], states_goal: List[S]) -> List[G]:
        """ Given a state, return a goal that represents a set of goal states of which the given state is a member.
        Does not have to always return the same goal.

        @param states_start: List of start states
        @param states_goal List of states from which goals will be sampled
        @return: Goals
        """
        pass

    @abstractmethod
    def is_solved(self, states: List[S], goals: List[G]) -> List[bool]:
        """ Returns true if the state is a member of the set of goal states represented by the goal

        @param states: List of states
        @param goals: List of goals
        @return: List of booleans where the element at index i corresponds to whether or not the
        state at index i is a member of the set of goal states represented by the goal at index i
        """
        pass

    def get_start_goal_pairs(self, num_steps_l: List[int],
                             times: Optional[Times] = None) -> Tuple[List[S], List[G]]:
        # Initialize
        if times is None:
            times = Times()

        # Start states
        start_time = time.time()
        states_start: List[S] = self.get_start_states(len(num_steps_l))
        times.record_time("get_start_states", time.time() - start_time)

        # random walk
        start_time = time.time()
        states_goal: List[S] = self._random_walk(states_start, num_steps_l)
        times.record_time("random_walk", time.time() - start_time)

        # state to goal
        start_time = time.time()
        goals: List[G] = self.sample_goal(states_start, states_goal)
        times.record_time("sample_goal", time.time() - start_time)

        return states_start, goals

    def expand(self, states: List[S]) -> Tuple[List[List[S]], List[List[A]], List[List[float]]]:
        """ Generate all children for the state, assumes there is at least one child state
        @param states: List of states
        @return: Children of each state, actions, transition costs for each state
        """
        # TODO further validate
        # initialize
        states_exp_l: List[List[S]] = [[] for _ in range(len(states))]
        actions_exp_l: List[List[A]] = [[] for _ in range(len(states))]
        tcs_l: List[List[float]] = [[] for _ in range(len(states))]
        state_actions: List[List[A]] = self.get_state_actions(states)

        num_actions_tot: NDArray[np.int_] = np.array([len(x) for x in state_actions])
        num_actions_taken: NDArray[np.int_] = np.zeros(len(states), dtype=int)
        actions_lt: NDArray[np.bool_] = num_actions_taken < num_actions_tot

        # for each move, get next states, transition costs, and if solved
        while np.any(actions_lt):
            idxs: NDArray[np.int_] = np.where(actions_lt)[0]
            states_idxs: List[S] = [states[idx] for idx in idxs]
            actions_idxs: List[A] = [state_actions[idx].pop(0) for idx in idxs]

            # next state
            states_next, tcs_move = self.next_state(states_idxs, actions_idxs)

            # transition cost
            for exp_idx, idx in enumerate(idxs):
                states_exp_l[idx].append(states_next[exp_idx])
                actions_exp_l[idx].append(actions_idxs[exp_idx])
                tcs_l[idx].append(tcs_move[exp_idx])

            num_actions_taken[idxs] = num_actions_taken[idxs] + 1
            actions_lt[idxs] = num_actions_taken[idxs] < num_actions_tot[idxs]

        return states_exp_l, actions_exp_l, tcs_l

    @abstractmethod
    def states_goals_to_nnet_input(self, states: List[S], goals: List[G]) -> List[NDArray[Any]]:
        """ States and goals to numpy arrays to be fed to the neural network

        @param states: List of states
        @param goals: List of goals
        @return: List of numpy arrays. Each index along the first dimension of each array corresponds to the
        index of a state.
        """
        pass

    @abstractmethod
    def get_v_nnet(self) -> HeurFnNNet:
        """ Get the neural network model for value iteration and A* search

        @return: neural network model
        """
        pass

    @abstractmethod
    def get_q_nnet(self) -> HeurFnNNet:
        """ Get the neural network model for Q-learning and Q* search. Has not been implemented yet,
        so do not have to implement (raise NotImplementedError).

        @return: neural network model
        """
        pass

    @abstractmethod
    def get_pddl_domain(self) -> List[str]:
        """ Implement if using PDDL solvers, like fast-downward. Do not have to implement if not also using
        traiditional planners (raise NotImplementedError).

        :return:
        """
        pass

    @abstractmethod
    def state_goal_to_pddl_inst(self, state: S, goal: G) -> List[str]:
        """ Implement if using PDDL solvers, like fast-downward. Do not have to implement if not also using
        traiditional planners (raise NotImplementedError).

        :return:
        """
        pass

    @abstractmethod
    def pddl_action_to_action(self, pddl_action: str) -> int:
        """ Implement if using PDDL solvers, like fast-downward. Do not have to implement if not also using
        traiditional planners (raise NotImplementedError).

        :return:
        """
        pass

    @abstractmethod
    def visualize(self, states: Union[List[S], List[G]]) -> NDArray[np.float64]:
        """ Implement if visualizing states. If you are planning on visualizing states, you do not have to implement
        this (raise NotImplementedError).

        :return:
        """
        pass

    def _random_walk(self, states: List[S], num_steps_l: List[int]) -> List[S]:
        states_walk: List[S] = [state for state in states]

        num_steps: NDArray[np.int_] = np.array(num_steps_l)
        num_moves_curr: NDArray[np.int_] = np.zeros(len(states), dtype=int)
        moves_lt: NDArray[np.bool_] = num_moves_curr < num_steps
        while np.any(moves_lt):
            idxs: NDArray[np.int_] = np.where(moves_lt)[0]
            states_to_move = [states_walk[idx] for idx in idxs]

            states_moved, _ = self.next_state_rand(states_to_move)

            for move_idx, idx in enumerate(idxs):
                states_walk[idx] = states_moved[move_idx]

            num_moves_curr[idxs] = num_moves_curr[idxs] + 1

            moves_lt[idxs] = num_moves_curr[idxs] < num_steps[idxs]

        return states_walk


class EnvGrndAtoms(Environment[S, A, G]):
    def __init__(self, env_name: str):
        super().__init__(env_name)

    @abstractmethod
    def state_to_model(self, states: List[S]) -> List[Model]:
        pass

    @abstractmethod
    def model_to_state(self, models: List[Model]) -> List[S]:
        """ Assumes model is a fully specified state

        :param models:
        :return:
        """
        pass

    @abstractmethod
    def goal_to_model(self, goals: List[G]) -> List[Model]:
        pass

    @abstractmethod
    def model_to_goal(self, models: List[Model]) -> List[G]:
        pass

    def is_solved(self, states: List[S], goals: List[G]) -> List[bool]:
        """ Returns whether or not state is solved

        @param states: List of states
        @param goals: List of goals
        @return: Boolean numpy array where the element at index i corresponds to whether or not the
        state at index i is solved
        """
        models_g: List[Model] = self.goal_to_model(goals)
        is_solved_l: List[bool] = []
        models_s: List[Model] = self.state_to_model(states)
        for model_state, model_goal in zip(models_s, models_g):
            is_solved_l.append(model_goal.issubset(model_state))

        return is_solved_l

    def sample_goal(self, states_start: List[S], states_goal: List[S]) -> List[G]:
        models_g: List[Model] = []

        models_s: List[Model] = self.state_to_model(states_goal)
        keep_probs: NDArray[np.float64] = np.random.rand(len(states_goal))
        for model_s, keep_prob in zip(models_s, keep_probs):
            rand_subset: Set[Atom] = misc_utils.random_subset(model_s, keep_prob)
            models_g.append(frozenset(rand_subset))

        return self.model_to_goal(models_g)

    @abstractmethod
    def get_bk(self) -> List[str]:
        """ get background, each element in list is a line

        :return:
        """
        pass

    @abstractmethod
    def get_ground_atoms(self) -> List[Atom]:
        """ Get all possible ground atoms that can be used to make a state

        :return:
        """
        pass

    @abstractmethod
    def on_model(self, m) -> Model:
        """ Process results from clingo

        :param m:
        :return:
        """
        pass

    @abstractmethod
    def start_state_fixed(self, states: List[S]) -> List[Model]:
        """ Given the start state, what must also be true for the goal state (i.e. immovable walls)

        :param states:
        :return:
        """
        pass



from typing import List, Any

import sys

from multiprocessing import Queue
import queue
import os
import shutil


class Logger(object):
    def __init__(self, filename: str, mode: str = "a", echo: bool = True):
        self.terminal = sys.stdout
        self.log = open(filename, mode)
        self.echo: bool = echo

    def write(self, message):
        if self.echo:
            self.terminal.write(message)
        self.log.write(message)
        self.log.flush()

    def flush(self):
        pass


def get_nowait_noerr(q: Queue) -> Any:
    try:
        q_ret: Any = q.get_nowait()
        return q_ret
    except queue.Empty:
        return None


def get_while_not_empty(q: Queue) -> List[Any]:
    q_rets: List[Any] = []

    while not q.empty():
        try:
            q_ret: Any = q.get_nowait()
            q_rets.append(q_ret)
        except queue.Empty:
            break

    return q_rets


def get_in_order(q: Queue, num: int) -> List[Any]:
    ret_vals: List[Any] = [None for _ in range(num)]
    for _ in range(num):
        idx, val = q.get()
        ret_vals[idx] = val
    return ret_vals


def copy_dir_files(src_dir: str, dest_dir: str):
    src_files: List[str] = os.listdir(src_dir)
    for file_name in src_files:
        full_file_name: str = os.path.join(src_dir, file_name)
        if os.path.isfile(full_file_name):
            shutil.copy(full_file_name, dest_dir)
